/* dns.h of iplog v1.1 -- eric@ojnk.nu */
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <netdb.h>

#define DNS_TIME	1800
#define DNS_MAXSIZE	32
#define DNS_HASH(x)	(((x >> 24) + (x >> 16)) % (DNS_MAXSIZE + 1))

struct dns_cache {
	unsigned long address;
	char *host;
	time_t when;
};

typedef struct dns_cache dns_entry;
dns_entry table[DNS_MAXSIZE];

void init_dns_cache(void) {
	int n;

	for (n = 0; n <= DNS_MAXSIZE; n++) {
		table[n].address = (-1L);
		table[n].when = (-1L);
	}
}

unsigned char *get_dns_cache(u_long addr) {
	if (table[DNS_HASH(addr)].address == addr)
		return (table[DNS_HASH(addr)].host);

	add_dns_cache(addr);

	if (table[DNS_HASH(addr)].address == addr)
		return (table[DNS_HASH(addr)].host);
	else
		return NULL;
}

void add_dns_cache(u_long addr) {
	struct in_addr i;
	struct hostent *he;
	
	i.s_addr = addr;
	he = gethostbyaddr((char *)&i, sizeof(struct in_addr), AF_INET);

	if (he != NULL) {
		table[DNS_HASH(addr)].address = addr;
		table[DNS_HASH(addr)].host = strdup(he->h_name);
		time(&table[DNS_HASH(addr)].when);
	}
}

void expire_dns() {
	time_t tm;
	int n;

	time(&tm);

	for (n = 0; n <= DNS_MAXSIZE; n++)
		if (table[n].address != (-1L))
			if (tm >= (table[n].when + DNS_TIME)) {
				free(table[n].host);
				table[n].address = (-1L);
				table[n].when = (-1L);
			}
}

