/* icmplog.c of iplog v1.1 -- eric@ojnk.nu */
#include <stdio.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <signal.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#define __FAVOR_BSD
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/ip_icmp.h>
#include <arpa/inet.h>

#include "iplog.h"
#include "lockfile.h"

#define icmp_len	sizeof(struct ip)+sizeof(struct icmphdr)

static u_char *protolookup(int);

struct packet {
	struct ip ip;
	struct icmphdr icmp;
	struct ip ipl;
	struct tcphdr tcp;
};

static const char *codes[] = {
	"echo reply from %s sz=%u(+%u)  [seq=0x%x]", /* ICMP_ECHOREPLY */
	"undefined icmp type 1 recieved from %s(?)", /* ? */
	"undefined icmp type 2 recieved from %s(?)", /* ? */
	"destination unreachable from %s", /* ICMP_DEST_UNREACH */
	"source quench from %s", /* ICMP_SOURCE_QUENCH */
	"redirect message from %s", /* ICMP_REDIRECT */
	"undefined icmp type 6 recieved from %s(?)", /* ? */
	"undefined icmp type 7 recieved from %s(?)", /* ? */
	"echo from %s sz=%u(+%u)  [seq=0x%x]",	/* ICMP_ECHO */
	"router advertisement from %s", /* ICMP_ROUTERADVERT (rfc1256) */
	"router solicitation from %s", /* ICMP_ROUTERSOLICIT (rfc1256) */
	"%s: time exceeded", /* ICMP_TIME_EXCEEDED */
	"parameter problem with %s", /* ICMP_PARAMETERPROB */
	"timestamp request from %s", /* ICMP_TIMESTAMP */
	"timestamp reply from %s", /* ICMP_TIMESTAMPREPLY */
	"info request from %s", /* ICMP_INFO_REQUEST */
	"info reply from %s", /* ICMP_INFO_REPLY */
	"address mask request from %s", /* ICMP_ADDRESS */
	"address mask reply from %s" /* ICMP_ADDRESSREPLY */
};

static const char *unreach[] = {
"%s: network is unreachable [%s: dp=%u sp=%u seq=0x%x]", /* ICMP_NET_UNREACH */
"%s: host is unreachable [%s: dp=%u sp=%u seq=0x%x]", /* ICMP_HOST_UNREACH */
	"%s: %s: protocol is unreachable", /* ICMP_PROT_UNREACH */
	"%s: %s port is unreachable [dp=%u sp=%u seq=0x%x]", /* ICMP_PORT_UNREACH */
	"%s: fragmentation needed", /* ICMP_FRAG_NEEDED */
	"%s: source route failed [%s: dp=%u sp=%u seq=0x%x]", /* ICMP_SR_FAILED */
	"%s: network unreachable [%s: dp=%u sp=%u seq=0x%x]", /* ICMP_NET_UNKNOWN */
	"%s: host unreachable [%s: dp=%u sp=%u seq=0x%x]", /* ICMP_HOST_UNKNOWN */
	"%s: host is isolated from source", /* ICMP_HOST_ISOLATED */
	"%s: network is unreachable (ANO) ", /* ICMP_NET_ANO */
	"%s: host is unreachable (ANO)", /* ICMP_HOST_ANO */
	"%s: network is unreachable (TOS)", /* ICMP_NET_UNR_TOS */
	"%s: host is unreachable (TOS)", /* ICMP_HOST_UNR_TOS */
	"%s: packet filtered [%s: dp=%u sp=%u seq=0x%x]", /* ICMP_PKT_FILTERED */
	"%s: precedence violation", /* ICMP_PREC_VIOLATION */
	"%s: precedence cutoff" /* ICMP_PREC_CUTOFF */
};

static const char *redir[] = {
	"%s: redirect [net]", /* ICMP_REDIR_NET */
	"%s: redirect [host]", /* ICMP_REDIR_HOST */
	"%s: redirect [net] (TOS)", /* ICMP_REDIR_NETTOS */
	"%s: redirect [host] (TOS)" /* ICMP_REDIR_HOSTTOS */
};

int main(void) {
	int sock;
	struct packet pkt;
	struct in_addr in;

	if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0) {
		perror("socket");
		exit(-1);
	}

	openlog("icmplog", LOGGING_SEVERITY, LOGGING_FACILITY);
	signal(SIGHUP, SIG_IGN);

	fork_to_background();
	initialize_lockfile("icmplog");
	drop_privileges();
	
	for (;;) {
		read(sock, (struct packet *) &pkt, sizeof(pkt));

	 	in.s_addr=pkt.ip.ip_src.s_addr;

		if (pkt.ip.ip_hl != 5) {
			syslog(LEVEL, "suspicious ip options from %s", inet_ntoa(in));
			continue;
		}

		switch (pkt.icmp.type) {
			case ICMP_TIME_EXCEEDED:
				break; /* ignored, because traceroute(1) will make logs messy */
			case ICMP_DEST_UNREACH:
				syslog(LEVEL, unreach[(int)pkt.icmp.code],inet_ntoa(in),							protolookup(pkt.ipl.ip_p),														ntohs(pkt.tcp.th_dport), ntohs(pkt.tcp.th_sport),								ntohs(pkt.tcp.th_seq));
					break;
			case ICMP_REDIRECT:
				syslog(LEVEL, redir[(int)pkt.icmp.code], inet_ntoa(in));
				break;
			case ICMP_ECHO:
			case ICMP_ECHOREPLY:
				syslog(LEVEL, codes[(int)pkt.icmp.type], inet_ntoa(in),								ntohs(pkt.ip.ip_len - htons((int)icmp_len)) , icmp_len,							ntohs(pkt.icmp.un.echo.sequence),
					ntohs(pkt.icmp.un.echo.id));
				break;
			default:
				syslog(LEVEL, codes[(int)pkt.icmp.type], inet_ntoa(in));
				break;
		}
	}
}

static u_char *protolookup(int num) {
	struct protoent *pe;
	static char buff[16];

	if ((pe = getprotobynumber(num)) != NULL)
		return (pe->p_name);
	else {
		sprintf(buff, "proto %d", num);
		return (buff);
	}
}
