/* lockfile.h from iplog v1.1 -- eric@ojnk.nu */
#include <stdlib.h>
#include <malloc.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <paths.h>
#include <errno.h>
extern int errno;


#ifndef _PATH_VARRUN
#	define	_PATH_VARRUN	"/var/run/"
#endif

char *my_getpid(void) {
	static char pid[10];

	sprintf(pid, "%d\n", getpid());
	return (pid);
}

int initialize_lockfile(char *p) {
	struct stat st;
	char *name, buf[8];
	int fd = 0, pid = 0;

	name = (char *)malloc(strlen(_PATH_VARRUN) + strlen(p) + 5);

	if (name == NULL)
		return (-1);

	sprintf(name, "%s%s.pid", _PATH_VARRUN, p);

	if (lstat(name, &st) != (-1L)) {
		if (S_ISREG(st.st_mode)) {
			if ((fd = open(name, O_RDONLY)) != (-1L)) {
				if (read(fd, buf, 8) > 0)
					sscanf(buf, "%d\n", &pid);

				if (kill(pid, 0) && (errno == ESRCH)) {
					if (unlink(name) != (-1L)) {
						syslog(LEVEL, "removed stale lockfile from pid %d",pid);
						(void)close(fd);
						goto create_lockfile;
					} else {
						syslog(LEVEL, "couldn't remove stale lockfile %s",name);
						(void)close(fd);
						exit(-1);
					}
				} else if (flock(fd, LOCK_EX|LOCK_NB) == (-1L)) {
					syslog(LEVEL, "%s is already running (pid: %d)", p, pid);
					(void)close(fd);
					exit(-1);
				}
			}
		} else {
			syslog(LEVEL, "lockfile %s isn't a regular file", name);
			(void)close(fd);
			exit(-1);
		}
	}

create_lockfile:

	if ((fd = open(name, O_RDWR|O_CREAT|O_TRUNC, 0644)) != (-1L)) {
		write(fd, my_getpid(), strlen(my_getpid()));

		if (flock(fd, LOCK_EX|LOCK_NB) != (-1L)) {
			free(name);
			return (1);
		} else {
			syslog(LEVEL, "lockfile not initialized (%m)");
			(void)close(fd);
			exit(-1);
		}
	}
	return (0);
}
