/* tcplog.c of iplog v1.1 -- eric@ojnk.nu */
#include <unistd.h>
#include <time.h>
#include <syslog.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#define __FAVOR_BSD
#include <netinet/tcp.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netdb.h>
#include <errno.h>
#include <ctype.h>
#include <arpa/inet.h>

#include "iplog.h"
#include "lockfile.h"

#ifdef RETRIEVE_IDENTD
u_char *get_identd_data(int, int, u_long, u_long);
int l_port(u_short);
#include "ident.h"
#endif

#ifdef USE_DNS_CACHE
u_char *get_dns_cache(u_long);
void add_dns_cache(u_long);
void expire_dns();
#include "dns.h"
#endif

static u_char *servlookup(u_short);
static u_char *hostlookup(u_long, int);

struct header {
	struct ip ip;
	struct tcphdr tcp;
};


#ifdef USE_DNS_CACHE
#	define SYN_FLOOD	12
#else
#	define SYN_FLOOD	8
#endif

#define FIN_LOGGED	0x01
#define NULL_LOGGED	0x02

#ifdef DETECT_BOGUS
#define TH_BOGUS	0x40
#define TH_BOG		TH_BOGUS
#endif

#ifdef RETRIEVE_IDENTD
void sigchld(int stat) {
	while (waitpid(-1, &stat, WNOHANG) > 0);
	signal(SIGCHLD, sigchld);
	return;
}
#endif

#ifdef USE_DNS_CACHE
void sighup(int stat) {
	syslog(LEVEL, "caught hangup signal, wiping dns cache");
	init_dns_cache();
	signal(SIGHUP, sighup);
	return;
}
#endif

int main(void) {
	int sock;
	time_t tm;
	struct header hdr;
#if defined(DETECT_FIN) || defined(DETECT_NULL)
	unsigned short flags = 0;
#endif
#ifdef DETECT_FIN
	time_t finlast;
	register u_char fincount = 0;
	u_long fin_last = 0;
#endif
#ifdef DETECT_NULL
	time_t nulllast;
	u_long null_last = 0;
#endif
#ifdef DETECT_BOGUS
	u_long last_bogus = 0;
#endif
#ifdef RETRIEVE_IDENTD
	char user[64] = "";
#endif
	time_t synlast;
	register u_char syncount = 0;

	if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0) {
		perror("socket");
		exit(-1);
	}
	openlog("tcplog", LOGGING_SEVERITY, LOGGING_FACILITY);
#ifdef RETRIEVE_IDENTD
	signal(SIGCHLD, sigchld);
#endif
#ifdef USE_DNS_CACHE
	signal(SIGHUP, sighup);
	init_dns_cache();
#else
	signal(SIGHUP, SIG_IGN);
#endif
	fork_to_background();
	initialize_lockfile("tcplog");
	drop_privileges();

	for (;;) {
		read(sock, (struct header *)&hdr, sizeof(hdr));
		time(&tm);

		if (syncount) {
			syncount -= ((tm - synlast) >> 1);	
			time(&synlast);
		}

#ifdef DETECT_BOGUS
		/* Nmap and Queso use a bogus tcp flag to "fingerprint" OS'es.. */
		if ((hdr.tcp.th_flags & TH_BOG) && last_bogus != hdr.ip.ip_src.s_addr) {
			last_bogus = hdr.ip.ip_src.s_addr;
			syslog(LEVEL, "bogus tcp flags set by %s (%s)",											hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)),						inet_ntoa(hdr.ip.ip_src));
		}
#endif

#ifdef DETECT_NULL
		if (!hdr.tcp.th_flags) {
			if (tm >= (nulllast + 60))
				flags &= ~NULL_LOGGED;
			if ((hdr.ip.ip_src.s_addr == null_last) && !(flags & NULL_LOGGED)) {
				syslog(LEVEL, "null scan detected from %s (%s)",									hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)),						inet_ntoa(hdr.ip.ip_src));
				flags |= NULL_LOGGED;
				time(&nulllast);
			} else if (hdr.ip.ip_src.s_addr != null_last) {
				flags &= ~NULL_LOGGED;
				null_last = hdr.ip.ip_src.s_addr;
			}
			continue;
		}
#endif

#ifdef DETECT_FIN
		if ((hdr.tcp.th_flags & TH_FIN) && !(hdr.tcp.th_flags & TH_ACK)) {
			if (tm >= (finlast + 60))
				flags &= ~FIN_LOGGED;
			if ((hdr.ip.ip_src.s_addr == fin_last) && !(flags & FIN_LOGGED)) {
				if (++fincount >= 2) {
					syslog(LEVEL, "%s scan detected from %s (%s)",										(hdr.tcp.th_flags & TH_URG) ? "Xmas" : "fin",									hostlookup(hdr.ip.ip_src.s_addr, (syncount!=SYN_FLOOD)),						inet_ntoa(hdr.ip.ip_src));
					flags |= FIN_LOGGED;
					time(&finlast);
				}
			} else if (hdr.ip.ip_src.s_addr != fin_last) {
				flags &= ~FIN_LOGGED;
				fincount = 1;
				fin_last = hdr.ip.ip_src.s_addr;
			}
			continue;
		}
#endif
		if ((hdr.tcp.th_flags & TH_SYN) && !(hdr.tcp.th_flags & TH_ACK)) {
			time(&synlast);

			if (syncount != SYN_FLOOD)
				syncount++;

#ifndef LOG_FTP_TRANSFERS
			if (hdr.tcp.th_sport == htons(20) &&												ntohs(hdr.tcp.th_dport) >= 1024)
				continue;
#endif

#ifdef RETRIEVE_IDENTD
			if (hdr.tcp.th_dport != htons(AUTH_PORT) && (syncount != SYN_FLOOD)					&& l_port(ntohs(hdr.tcp.th_dport))) {
				if (!fork()) {
					memset(user, 0, 64);
					snprintf(user, 64, "%s",															get_identd_data(ntohs(hdr.tcp.th_sport),										ntohs(hdr.tcp.th_dport), hdr.ip.ip_src.s_addr,									hdr.ip.ip_dst.s_addr));

					syslog(LEVEL, "%s connection attempt from %s%s%s:%u",								servlookup(hdr.tcp.th_dport), user,												(user[0] ? "@" : ""),															hostlookup(hdr.ip.ip_src.s_addr, (syncount!=SYN_FLOOD)),						ntohs(hdr.tcp.th_sport));
					exit(0);
				}
			} else {
				syslog(LEVEL, "%s connection attempt from %s:%u",										servlookup(hdr.tcp.th_dport), 													hostlookup(hdr.ip.ip_src.s_addr, (syncount!=SYN_FLOOD)),						ntohs(hdr.tcp.th_sport));
			}
		}
#else				
			syslog(LEVEL, "%s connection attempt from %s:%u",										servlookup(hdr.tcp.th_dport),													hostlookup(hdr.ip.ip_src.s_addr, (syncount != SYN_FLOOD)),						ntohs(hdr.tcp.th_sport));
		}
#endif
	}
	(void)close(sock);
}

static u_char *servlookup(u_short port) {
	struct servent *se;
	static char buff[16];

	if ((se = getservbyport(port, "tcp")) != NULL)
		return (se->s_name);
	else {
		sprintf(buff, "port %u", ntohs(port));
		return buff;
	}
}

static u_char *hostlookup(u_long addr, int dohost) {
	struct in_addr i;
	i.s_addr = addr;

	if (!dohost) {
		return (inet_ntoa(i));
	} else {
#ifdef USE_DNS_CACHE
		register char *host;
		expire_dns();		

		if ((host = get_dns_cache(addr)) != NULL)
			return (host);
		else
			return (inet_ntoa(i));
	}
#else
		register struct hostent *he;
		he = gethostbyaddr((char *)&i, sizeof(struct in_addr), AF_INET);
		if (he == NULL)
			return (inet_ntoa(i));
		else
			return (he->h_name);
	}
#endif
	return ("unknown");
}
