/* udplog.c of iplog v1.1 -- eric@ojnk.nu */
#include <unistd.h>
#include <time.h>
#include <syslog.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <netinet/udp.h>
#include <netinet/ip.h>
#include <netinet/in.h>
#include <netdb.h>
#include <fcntl.h>
#include <ctype.h>
#include <arpa/inet.h>

#include "iplog.h"
#include "lockfile.h"

#define MAX 64 /* i think that's reasonable.. */

#ifndef _PATH_RESCONF /* bleh*/
#	define _PATH_RESCONF	"/etc/resolv.conf"
#endif

#ifdef DETECT_TRACEROUTE
int is_traceroute(u_short a, u_short b, int ttl) {
	return ((ttl == 1) && (ntohs(a) > 32000) && (ntohs(a) < 40000)							&& (ntohs(b) > 32000) && (ntohs(b) < 40000));
}
u_char *hostlookup(u_long);
#endif

u_char *servlookup(u_short);
u_long name_resolve(u_char *);
void reread_ns(int);
void readns(u_long *);

u_long ignore[MAX];

struct packet {
	struct ip ip;
	struct udphdr udp;
};

void readns(u_long *ignore) {
	FILE *fp;
	int found = 0;
	char line[BUFSIZ], *p = NULL;

	if ((fp = fopen(_PATH_RESCONF, "r")) != NULL) {
		while ((fgets(line, sizeof(line), fp) != NULL) && (found < MAX - 1)) {
			if ((p = strchr(line, '\n')) != NULL) {
				*p = '\0';
				if (!strncmp(line, "nameserver ", 11))
					if (name_resolve(1+(strrchr(line, ' '))) != (-1L))
						ignore[found++] = name_resolve(1+(strrchr(line, ' ')));
			}
		}		
		(void)fclose(fp);
		syslog(LEVEL, "cached %d nameserver%s", found, (found == 1 ? "" : "s"));
		ignore[found++] = (-1L);
	} else {
		perror(_PATH_RESCONF);
		exit(-1);
	}
}

void reread_ns(int unused) {
	syslog(LEVEL, "caught hangup signal, re-reading %s", _PATH_RESCONF);
	readns(ignore);
	signal(SIGHUP, reread_ns);
}

int main(void) {
	struct packet pkt;
	struct in_addr in;
	int sock, i, bad = 0;
#ifdef DETECT_TRACEROUTE
	time_t lasttr = 0;
	u_long last = 0;
#endif

	if ((sock = socket(AF_INET, SOCK_RAW, IPPROTO_UDP)) < 0) {
		perror("socket");
		exit(-1);
	}
	openlog("udplog", LOGGING_SEVERITY, LOGGING_FACILITY);
	signal(SIGHUP, reread_ns);

	fork_to_background();
	initialize_lockfile("udplog");
	drop_privileges();

	readns(ignore);

	for (;;) {
		read(sock, (struct header *) &pkt, sizeof(pkt));

		if (ntohs(pkt.udp.source) == 53) {
			for (i=0, bad=0; ignore[i] != (-1L); i++) {
				if (pkt.ip.ip_src.s_addr == ignore[i]) {
					bad = 1;
					break;
				}
			}
		} else bad = 0;

		if (!ntohs(pkt.udp.dest) || !ntohs(pkt.udp.source)) /* Odd stuff */
			bad = 1;

		if (!bad) {
			in.s_addr = pkt.ip.ip_src.s_addr;
#ifdef DETECT_TRACEROUTE
			if (is_traceroute(pkt.udp.dest, pkt.udp.source, pkt.ip.ip_ttl)) {
				if ((last != in.s_addr) || (time(NULL) <= (lasttr + 15))) {
					syslog(LEVEL, "traceroute from %s (%s)",												hostlookup(in.s_addr), inet_ntoa(in));
					time(&lasttr);
					last = in.s_addr;
				}
			} else
#endif
			syslog(LEVEL, "dgram to %s from %s:%u sz=%u(+%u)",									servlookup(pkt.udp.dest), inet_ntoa(in),										ntohs(pkt.udp.source),															ntohs(pkt.ip.ip_len) - sizeof(pkt.ip) - sizeof(pkt.udp),						sizeof(pkt.ip) + sizeof(pkt.udp));
		}
	}
	close(sock);
}

u_char *servlookup(u_short port) {
	struct servent *se;
	static u_char buff[16];
	if ((se = getservbyport(port, "udp")) != NULL)
		return (se->s_name);
	else {
		sprintf(buff, "port %u", ntohs(port));
		return (buff);
	}
}

#ifdef DETECT_TRACEROUTE
u_char *hostlookup(u_long addr) {
	register struct hostent *he;
	struct in_addr i;

	i.s_addr = addr;

	he = gethostbyaddr((char *)&i, sizeof(struct in_addr), AF_INET);
	if (he == NULL)
		return (inet_ntoa(i));
	else
		return (he->h_name);

	return ("unknown");
}
#endif

/* name_resolve function from route's libnet */
u_long name_resolve(u_char *host_name) {
	u_long l;
	u_int val;
	int i;

	if (!isdigit(host_name[0]))
		return (-1L);

	l = 0;
	for (i = 0; i < 4; i++) {
		l <<= 8;
		if (*host_name) {
			val = 0;

			while (*host_name && *host_name != '.') {
				val *= 10;
				val += *host_name - '0';
				host_name++;
			}

			l |= val;
			if (*host_name) host_name++;
		}
	}
	return (htonl(l));
}
